@@ -34,6 +34,8 @@ import inspect
 import glob
 import json
 from urllib.parse import urlparse
+import string
+import random
 
 # import astpretty
 
@@ -81,7 +83,7 @@ def get_label_info(label):
         'Content-Type': 'application/x-www-form-urlencoded',
         'Authorization': token
         }
-    url = 'https://mlflow.' + service + '.com/api/2.0/mlflow/label/get?labelname=' + label
+    url = 'https://mlflow.' + service + '/api/2.0/mlflow/label/get?labelname=' + label
     try:
         response = requests.post(url, headers=headers)
         response.raise_for_status()
@@ -102,7 +104,7 @@ def get_xform_info(xformname):
         'Content-Type': 'application/x-www-form-urlencoded',
         'Authorization': token
         }
-    url = 'https://mlflow.' + service + '.com/api/2.0/mlflow/xform/get?transform_name=' + xformname
+    url = 'https://mlflow.' + service + '/api/2.0/mlflow/xform/get?transform_name=' + xformname
     try:
         response = requests.get(url, headers=headers)
         response.raise_for_status()
@@ -120,7 +122,7 @@ def get_xform_info(xformname):
 
 def download_one_dir(label):
     timespec, bucketname, prefix, service = get_label_info(label)
-    endpoint = "https://lbl" + label + ".s3proxy." + service + ".com:443/";
+    endpoint = "https://lbl" + label + ".s3proxy." + service + ":443/";
 
     session = boto3.session.Session(profile_name='infinstor')
     client = session.client('s3', endpoint_url=endpoint)
@@ -509,7 +511,7 @@ def actually_run_transformation(client, is_pandas_df, bucketname,\
         status_str = 'Success'
     # print('Globals=')
     # pprint.pprint(glb)
-    info("Execution of global statics complete. status_str=" + status_str)
+    # info("Execution of global statics complete. status_str=" + status_str)
 
     namespaced_infin_transform_fnx = namespaced_function(glb[xn], glb, None, True)
     try:
@@ -530,8 +532,33 @@ def actually_run_transformation(client, is_pandas_df, bucketname,\
         print('saving tf.data.Dataset unimplemented')
     return objects
 
+def run_script(transformation_string, **kwargs):
+    transformSrc = transformation_string
+    if (verbose == True):
+        print('run_script: transformSrc=' + transformSrc);
+
+    tree = ast.parse(transformSrc)
+
+    compiledcode2 = compile(tree, "<string>", "exec")
+
+    # Add all functions in xformcode to the globals dictionary
+    glb = {}
+    fl = FuncLister(glb)
+    fl.visit(tree)
+    try:
+        exec(compiledcode2, glb)
+    except Exception as e:
+        status_str = str(e)
+        info("Execution of global statics failed. status_str=" + status_str)
+        raise
+    else:
+        status_str = 'Success'
+    #print('Globals=')
+    #pprint.pprint(glb)
+    #info("Execution of script complete. status_str=" + status_str)
+
 def run_xform_periodically(seconds, service_name, bucketname, prefix_in, xformname, **kwargs):
-    endpoint = "https://s3proxy." + service_name + ".com:443/";
+    endpoint = "https://s3proxy." + service_name + ":443/";
     print('infinstor proxy endpoint=' + endpoint)
     session = boto3.session.Session(profile_name='infinstor')
     client = session.client('s3', endpoint_url=endpoint)
@@ -806,6 +833,9 @@ def generate_cache_key(input_data_spec, xformname):
         cache_key = 'xform/' + xformname + '/label/' + input_data_spec['label']
     elif (input_data_spec['type'] == 'mlflow-run-artifacts'):
         cache_key = 'xform/' + xformname + '/run_id/' + input_data_spec['run_id']
+    elif (input_data_spec['type'] == 'no-input-data'):
+        cache_key = 'xform/' + xformname + '/no-input-data/' \
+            +''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(8))
     else:
         raise ValueError('Error. Unknown input_data_spec type ' + input_data_spec['type'])
     return cache_key
@@ -821,6 +851,8 @@ def generate_input_data_spec_string(input_data_spec):
         rv = 'label/' + input_data_spec['label']
     elif (input_data_spec['type'] == 'mlflow-run-artifacts'):
         rv = 'run_id/' + input_data_spec['run_id']
+    elif (input_data_spec['type'] == 'no-input-data'):
+        rv = 'no-input-data'
     else:
         raise ValueError('Error. Unknown input_data_spec type ' + input_data_spec['type'])
     return rv
@@ -845,7 +877,7 @@ def run_transform_inline(service_name, run_id, input_data_spec, xformname, **kwa
     cache_key = generate_cache_key(input_data_spec, xformname)
     if (input_data_spec['type'] == 'infinsnap' or input_data_spec['type'] == 'infinslice'):
         endpoint = "https://" + input_data_spec['time_spec'] + ".s3proxy."\
-                + service_name + ".com:443/";
+                + service_name + ":443/";
         prefix_in = input_data_spec['prefix']
         prefix_trunc = ''
         bucketname = input_data_spec['bucketname']
@@ -853,7 +885,7 @@ def run_transform_inline(service_name, run_id, input_data_spec, xformname, **kwa
         client = session.client('s3', endpoint_url=endpoint)
     elif (input_data_spec['type'] == 'label'):
         endpoint = "https://lbl" + input_data_spec['label'] + ".s3proxy."\
-                + service_name + ".com:443/";
+                + service_name + ":443/";
         prefix_in = '' # ignored for labels
         prefix_trunc = '' # ignored for labels
         bucketname = '' # ignored for labels
@@ -864,6 +896,13 @@ def run_transform_inline(service_name, run_id, input_data_spec, xformname, **kwa
         prefix_trunc = prefix_in
         session = boto3.session.Session()
         client = session.client('s3')
+    elif (input_data_spec['type'] == 'no-input-data'):
+        endpoint = None
+        prefix_in = None
+        prefix_trunc = ''
+        bucketname = None
+        session = None
+        client = None
     else:
         raise ValueError('Error. Unknown input_data_spec type ' + input_data_spec['type'])
 
@@ -881,39 +920,53 @@ def run_transform_inline(service_name, run_id, input_data_spec, xformname, **kwa
 def run_transform_string_inline(client, bucketname, prefix_in, prefix_trunc,
         transform_string, **kwargs):
     if (look_for_transform(transform_string, TRANSFORM_RAW_PD)):
+        if (client == None):
+            raise ValueError('Error. Transform type ' + str(TRANSFORM_RAW_PD) + ' needs input_data')
         print("Running transform " + str(TRANSFORM_RAW_PD))
         return actually_run_transformation(client, True,\
                 bucketname, prefix_in, prefix_trunc, False, transform_string,\
                 **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_RAW_DS)):
+        if (client == None):
+            raise ValueError('Error. Transform type ' + str(TRANSFORM_RAW_DS) + ' needs input_data')
         print("Running transform " + str(TRANSFORM_RAW_DS))
         return actually_run_transformation(client, False,\
                 bucketname, prefix_in, prefix_trunc, False, transform_string,\
                 **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_CSV_PD)):
+        if (client == None):
+            raise ValueError('Error. Transform type ' + str(TRANSFORM_CSV_PD) + ' needs input_data')
         print("Running transform " + str(TRANSFORM_CSV_PD))
         return actually_run_transformation(client, True,\
                 bucketname, prefix_in, prefix_trunc, True, transform_string,\
                 **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_CSV_DS)):
+        if (client == None):
+            raise ValueError('Error. Transform type ' + str(TRANSFORM_CSV_DS) + ' needs input_data')
         print("Running transform " + str(TRANSFORM_CSV_DS))
         return actually_run_transformation(client, False,\
                 bucketname, prefix_in, prefix_trunc, True, transform_string,\
                 **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_ONE_OBJ)):
+        if (client == None):
+            raise ValueError('Error. Transform type ' + str(TRANSFORM_ONE_OBJ) + ' needs input_data')
         print("Running transform " + str(TRANSFORM_ONE_OBJ))
         return read_and_xform_one_object(client, bucketname, prefix_in, prefix_trunc,
                 transform_string, **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_DIR_BY_DIR)):
+        if (client == None):
+            raise ValueError('Error. Transform type ' + str(TRANSFORM_DIR_BY_DIR) + ' needs input_data')
         print("Running transform " + str(TRANSFORM_DIR_BY_DIR))
         return read_and_xform_dir_by_dir(client, bucketname, prefix_in, prefix_trunc,
                 transform_string, **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_ALL_OBJECTS)):
+        if (client == None):
+            raise ValueError('Error. Transform type ' + str(TRANSFORM_ALL_OBJECTS) + ' needs input_data')
         print("Running transform " + str(TRANSFORM_ALL_OBJECTS))
         return read_and_xform_all_objects(client, bucketname, prefix_in, prefix_trunc,
                 transform_string, **kwargs)
     else:
-        raise Exception("Cannot find a known transform function");
+        run_script(transform_string, **kwargs)
 
 def run_transform_singlevm(service_name, parent_run_id, last_in_chain_of_xforms,
         input_data_spec, xformname, instance_type, **kwargs):
@@ -1116,7 +1169,7 @@ def test_infin_transform(service_name, input_data_spec, **kwargs):
 
     if (input_data_spec['type'] == 'infinsnap' or input_data_spec['type'] == 'infinslice'):
         endpoint = "https://" + input_data_spec['time_spec'] + ".s3proxy."\
-                + service_name + ".com:443/";
+                + service_name + ":443/";
         prefix_in = input_data_spec['prefix']
         prefix_trunc = ''
         bucketname = input_data_spec['bucketname']
@@ -1124,7 +1177,7 @@ def test_infin_transform(service_name, input_data_spec, **kwargs):
         client = session.client('s3', endpoint_url=endpoint)
     elif (input_data_spec['type'] == 'label'):
         endpoint = "https://lbl" + input_data_spec['label'] + ".s3proxy."\
-                + service_name + ".com:443/";
+                + service_name + ":443/";
         prefix_in = '' # ignored for labels
         prefix_trunc = '' # ignored for labels
         bucketname = '' # ignored for labels
@@ -1135,6 +1188,13 @@ def test_infin_transform(service_name, input_data_spec, **kwargs):
         prefix_trunc = prefix_in
         session = boto3.session.Session()
         client = session.client('s3')
+    elif (input_data_spec['type'] == 'no-input-data'):
+        endpoint = None
+        prefix_in = None
+        prefix_trunc = ''
+        bucketname = None
+        session = None
+        client = None
     else:
         raise ValueError('Error. Unknown input_data_spec type ' + input_data_spec['type'])
 
